package com.pwt.netty;

import cn.hutool.core.lang.UUID;
import cn.hutool.json.JSONUtil;
import com.pwt.common.constant.LogConstants;
import com.pwt.common.utils.ProxyUtils;
import io.netty.bootstrap.Bootstrap;
import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.*;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.channel.socket.nio.NioSocketChannel;
import io.netty.handler.codec.bytes.ByteArrayDecoder;
import io.netty.handler.codec.bytes.ByteArrayEncoder;
import io.netty.handler.logging.LogLevel;
import io.netty.handler.logging.LoggingHandler;
import io.netty.util.concurrent.DefaultThreadFactory;
import io.netty.util.internal.SystemPropertyUtil;
import lombok.extern.slf4j.Slf4j;

import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;

@Slf4j(topic = LogConstants.PROXY_BIZ)
public class ProxyServerNetty {
    private static final int DEFAULT_EVENT_LOOP_THREADS = Math.max(1,
            SystemPropertyUtil.getInt("io.netty.eventLoopThreads", Runtime.getRuntime().availableProcessors() * 2));
//
    private final List<String> proxyIpPortList;
    private final int port;
    private final String host;

    private final Map<String, Channel> clientChannelMap = new ConcurrentHashMap<>();
    private final Map<String, List<Channel>> targetChannelsMap = new ConcurrentHashMap<>();

    private EventLoopGroup bossGroup;
    private EventLoopGroup workerGroup;

    private boolean isRunning;

    public ProxyServerNetty(List<String> proxyIpPortList, int port, String host) {
        this.proxyIpPortList = proxyIpPortList;
        this.port = port;
        this.host = host;
    }

    public void start() {
        bossGroup = new NioEventLoopGroup(1, new DefaultThreadFactory("proxyBoss"));
        workerGroup = new NioEventLoopGroup(DEFAULT_EVENT_LOOP_THREADS, new DefaultThreadFactory("proxyWorker"));

        ServerBootstrap b = new ServerBootstrap();
        b.group(bossGroup, workerGroup)
                .channel(NioServerSocketChannel.class)
                .handler(new LoggingHandler(LogLevel.INFO))
                .childHandler(new ChannelInitializer<SocketChannel>() {
                    @Override
                    protected void initChannel(SocketChannel ch) {
                        ch.pipeline().addLast(new ByteArrayDecoder(), new ByteArrayEncoder(), new ClientHandler());
                    }
                });

        log.info("开启代理服务");
        log.info("代理列表：{}", JSONUtil.toJsonStr(proxyIpPortList));

        ChannelFuture f = b.bind(new InetSocketAddress(host, port));
        f.syncUninterruptibly();
        f.channel();
    }

    public void stop() {
        log.info("关闭代理服务");
        try {
            if (bossGroup != null) {
                bossGroup.shutdownGracefully();
            }
            if (workerGroup != null) {
                workerGroup.shutdownGracefully();
            }
            for (Channel clientChannel : clientChannelMap.values()) {
                if (clientChannel != null) {
                    clientChannel.close();
                }
            }
            for (List<Channel> targetChannels : targetChannelsMap.values()) {
                if (targetChannels != null) {
                    for (Channel targetChannel : targetChannels) {
                        if (targetChannel != null) {
                            targetChannel.close();
                        }
                    }
                }
            }
        } catch (Exception e) {
            //
        }
    }

    class ClientHandler extends SimpleChannelInboundHandler<byte[]> {

        private String clientKey;
        private List<Channel> targetChannels;

        @Override
        public void channelActive(ChannelHandlerContext ctx) {
            log.info("接收到车辆连接：{}", ctx.channel().id().asLongText());
            clientKey = UUID.fastUUID().toString();
            clientChannelMap.put(clientKey, ctx.channel());

            targetChannels = new ArrayList<>(5);
            targetChannelsMap.put(clientKey, targetChannels);

            for (String proxyIpPort : proxyIpPortList) {
                connectToTarget(ctx, proxyIpPort);
            }
        }

        private void connectToTarget(ChannelHandlerContext ctx, String proxyIpPort) {
            String[] split = proxyIpPort.split(":");
            String ip = split[0];
            int port = Integer.parseInt(split[1]);

            Bootstrap b = new Bootstrap();
            b.group(ctx.channel().eventLoop())
                    .channel(NioSocketChannel.class)
                    .handler(new ChannelInitializer<SocketChannel>() {
                        @Override
                        protected void initChannel(SocketChannel ch) {
                            ch.pipeline().addLast(new ByteArrayDecoder(), new ByteArrayEncoder(), new TargetHandler(ctx.channel(), proxyIpPort));
                        }
                    });

            b.connect(ip, port).addListener((ChannelFutureListener) future -> {
                if (future.isSuccess()) {
                    targetChannels.add(future.channel());
                } else {
                    log.error("targetSocket 未成功连接: {}", proxyIpPort);
                    scheduleReconnect(ctx, proxyIpPort);
                }
            });
        }

        @Override
        protected void channelRead0(ChannelHandlerContext ctx, byte[] msg) {
            log.info("代理服务接收到车辆报文：{}", ProxyUtils.bytesToHex(msg));
            Iterator<Channel> iterator = targetChannels.iterator();
            while (iterator.hasNext()) {
                Channel targetChannel = iterator.next();
                if (!targetChannel.isOpen()) {
                    iterator.remove();
                    if (targetChannels.isEmpty()) {
                        log.info("所有 targetSocket 关闭，关闭 clientSocket");
                        ctx.close();
                    }
                    targetChannel.close();
                    continue;
                }
                if (targetChannel.isActive()) {
                    targetChannel.writeAndFlush(msg);
                }
            }
        }

        @Override
        public void channelInactive(ChannelHandlerContext ctx) {
            log.error("clientSocket 关闭，id: {}", ctx.channel() == null ? "" : ctx.channel().id().asLongText());
            for (Channel targetChannel : targetChannels) {
                if (targetChannel != null) {
                    targetChannel.close();
                }
            }
            targetChannelsMap.remove(clientKey);
            clientChannelMap.remove(clientKey);
            ctx.close();
            log.error("clientSocket 关闭成功");
        }

        @Override
        public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) {
            log.error("客户端信息解析失败！", cause);
            //ctx.close();
        }
    }

    class TargetHandler extends SimpleChannelInboundHandler<byte[]> {

        private final Channel clientChannel;
        private final String proxyIpPort;

        public TargetHandler(Channel clientChannel, String proxyIpPort) {
            this.clientChannel = clientChannel;
            this.proxyIpPort = proxyIpPort;
        }

        @Override
        protected void channelRead0(ChannelHandlerContext ctx, byte[] msg) {
            log.info("代理服务接收到换电站报文：{}", ProxyUtils.bytesToHex(msg));
            if (clientChannel.isActive()) {
                clientChannel.writeAndFlush(msg);
            }
        }

        @Override
        public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) {
            log.error("换电站信息解析失败！", cause);
            //ctx.close();
        }

        @Override
        public void channelInactive(ChannelHandlerContext ctx) {
            log.info("targetSocket 关闭");
            ctx.close();
        }
    }

    private void scheduleReconnect(ChannelHandlerContext ctx, String proxyIpPort) {
        ctx.channel().eventLoop().schedule(() -> {
            log.info("尝试重新连接到 targetSocket: {}", proxyIpPort);
            connectToTarget(ctx, proxyIpPort);
        }, 5, TimeUnit.SECONDS);
    }

    private void connectToTarget(ChannelHandlerContext ctx, String proxyIpPort) {
        String[] split = proxyIpPort.split(":");
        String ip = split[0];
        int port = Integer.parseInt(split[1]);

        Bootstrap b = new Bootstrap();
        b.group(ctx.channel().eventLoop())
                .channel(NioSocketChannel.class)
                .handler(new ChannelInitializer<SocketChannel>() {
                    @Override
                    protected void initChannel(SocketChannel ch) {
                        ch.pipeline().addLast(new ByteArrayDecoder(), new ByteArrayEncoder(), new TargetHandler(ctx.channel(), proxyIpPort));
                    }
                });

        b.connect(ip, port).addListener((ChannelFutureListener) future -> {
            if (future.isSuccess()) {
                log.info("成功重新连接到 targetSocket: {}", proxyIpPort);
                List<Channel> targetChannels = targetChannelsMap.get(ctx.channel().id().asLongText());
                if (targetChannels != null) {
                    targetChannels.add(future.channel());
                }
            } else {
                log.info("重新连接到 targetSocket 失败: {}", proxyIpPort);
                scheduleReconnect(ctx, proxyIpPort);
            }
        });
    }
}
